/*	fgrep is "field oriented grep
*	don't comfuse obsolete fgrep (fixed grep) in unix
*	fgrep is double times slower then Ken's genuine grep
*	we have another reason we need fgrep:
*	option "-i" of grep is only effective to ascii text.
*	probably it is not easy to fix the problem.
*	fgrep is OK
*	-Kenar-
*/

#include <u.h>
#include <libc.h>
#include <bio.h>
#include <regexp.h>
#include "misc.h"

#define MAXFIELD 256


Biobuf bout;
char *sep;
int fnum;
int gflag;
int hflag;
int iflag;
int nflag;
int sflag;
int vflag;
int mcount; // matched count

void
usage(void)
{
	fprint(2,"usage: fgrep [-ghinsv] [-f $num] [-F $sep] pattern [file ...]\n"
	" fgrep is a field oriented grep.\n"
	" don't confuse with obsolete fgrep that is the grep with -f option.\n"
	" -g	pattern is a glob pattern.\n"
	" -h	Do not print file name tags.\n"
	" -i	Ignore alphabetic case distinctions.\n"
	" -n	Mark each printed line with its line number.\n"
	" -s	Produce no output, but return status.\n"
	" -v	Reverse: print lines that do not match the pattern.\n"
	" -f $num	field number that begins with 1,2,...\n"
	" -F $sep	field separator\n"
	);
	exits("usage");
}

char *
gsub(char *src, char *fr, char *to)
/*	NB: fr is not a reg pattern	*/
{
	char *p,*s,*ep,*cp,*se;
	char *buf;
	char c;
	int size;
	size = strlen(src) + 256;
	cp = buf = malloc(size);
	ep = buf + size;
	s = src;
	se = s + strlen(src);
	*cp = 0;
	while(s < se){
		//print("s: %s\n",s);
		//print("b1: %s\n",buf);
		p = strstr(s,fr);
		if(!p) break;
		//print("p2: %s\n",p);
		c = *p;
		*p = 0;
		cp = strecpy(cp,ep,s);
		//print("b3: %s\n",buf);
		*p = c;
		cp = strecpy(cp,ep,to);
		//print("b4: %s\n",buf);
		s = p + strlen(fr);
	}
	strecpy(cp,ep,s);
	//print("b5: %s\n",buf);
	return buf; /* should be freed */
}



int
match(Reprog *re, char *str)
/*	test if matched	*/
{
	Resub m;
	m.sp = str;
	m.ep = nil;
	if(regexec(re, str, &m, 1))
		return 1;
	return 0;
}

void
doit(char *file, Reprog *re)
{
	char *p, *t, *buf, *args[MAXFIELD];
	Rdbuf *rb;
	int fd, n, m, mvalue;
	int bufsize = 1024; /* initial buf size */
	if(file == nil)
		fd = 0;
	else{
		fd = open(file,OREAD);
		if(fd < 0)
			sysfatal("%s not open: %r",file);
	}
	rb = rdinit(fd, 32*1024);
	buf = emalloc(bufsize);
	n = 1;
	mvalue = !vflag;
	while((p = rdline(rb))){
		if(rdlnlen(rb) > bufsize - 1){
			free(buf);
			bufsize = rdlnlen(rb) + 1024; /* at least + 1 */
			buf = emalloc(bufsize);
		}
		strecpy(buf, buf + bufsize, p);

		if(iflag)
			strtolower(buf);
		t = buf;
		if(fnum){
			if(sep)
				m = split(buf, args, MAXFIELD, *sep);
			else
				m = tokenize(buf, args, MAXFIELD);
			if(fnum > m)
				sysfatal("field is too few: %d %d",fnum,m);
			t = args[fnum-1];
		}
		if(match(re,t) == mvalue){
			if(sflag)
				exits(nil);
			if(nflag){
				if(file && !hflag)
					Bprint(&bout,"%s:%d: ", file, n);
				else
					Bprint(&bout,"%d: ", n);
			}
			Bprint(&bout,"%s\n", p);
			mcount++;
		}
		n++;
	}
	free(buf);
	rdfree(rb);
	if(fd)
		close(fd);
}


void
main(int argc, char *argv[])
{
	char *pat, *p0, *p1, *p2, *p3;
	Reprog *re;
	
	ARGBEGIN{
	case 'F': sep = EARGF(usage());
		break;
	case 'f': fnum = atoi(EARGF(usage()));
		break;
	case 'g': gflag++;
		break;
	case 'h': hflag++;
		break;
	case 'i': iflag++;
		break;
	case 'n': nflag++;
		break;
	case 's': sflag++;
		break;
	case 'v': vflag = 1; // don't vflag++;
		break;
	default: usage();
	}ARGEND

	if(*argv == nil)
		usage();
	pat = strdup(*argv++);
	if(gflag){
		/*	glob pattern
		*	"?", "*", "[~" in the pattern are glob symbols
		*	we assume some RE patterns such as "^", "$" are not in the pattern
		*	"(", ")", "|" are allowed. they act as "group", and "or".
		*	that is, our glob pattern is extended one.		*/
		
		p0 = gsub(pat,".","\.");
		p1 = gsub(p0,"?","."); free(p0);
		p2 = gsub(p1,"*",".*"); free(p1);
		p3 = gsub(p2,"[~","[^"); free(p2);
		pat = smprint("^(%s)$", p3); free(p3);
	}

	if(iflag)
		strtolower(pat);

	re = regcomp(pat);
	if(!re)
		sysfatal("garbaged expression %s",pat);

	Binit(&bout, 1, OWRITE);

	if(*argv == nil){
		doit(nil,re);
		goto Lend;
	}
	if(argc == 2)
		hflag++;
	while(*argv)
		doit(*argv++,re);
Lend:
	Bterm(&bout);
	free(re);
	free(pat);
	exits(mcount?nil:"no match");
}
