/*
*	8.lkdict Binit x
*/


#include <u.h>
#include <libc.h>
#include <bio.h>
#include <regexp.h>
#include "misc.h"

#define MAXFIELD 16
#define MAXBUF 1024
#define BUFSIZE 16*1024
#define TAB '\t'
#define SPACE ' '

//#define RDLIST_BIO
#define RDLIST_RAW
/*	RDLIST COMMENT
*	Which is better?
*	For very frequent term such as "th", I didn't observe difference between two methods.
*	The wc for Eijiro is,
*		283504 2688122 39223073
*	But "thema" makes difference. RDLIST_RAW is double time faster. The wc is,
*		44     234    4683
*	Conclusion: RDLIST_RAW is totally better than RDLIST_BIO
*/

#define DBG if(0)

typedef struct Pos Pos;
struct Pos{
	int count;
	vlong offset;
};

typedef struct Abuf Abuf;
struct Abuf {
	vlong *array;
	int size;
	int cp;
};

List *lroot[MAXFIELD];
char *prefix = "";
char *dir="";
int NoText;
char *outfmt="%lld\n";
int nflag;
char *header="";
char *nsep=":";

#ifdef RDLIST_BIO
Biobuf list;
#endif

int fd_indx, fd_dict, fd_list, fd_text;
char *comline;
Biobuf bout;


void
usage(char *msg)
{
	if(msg)
		fprint(2,"%s\n",msg);
	fprint(2,
	"usage: lkdict [-p $prefix] [-d $dir] [-H $header] [-S $nsep] [-Q] [-n] $query $dic [$text]\n"
	"lkdict look dictionary in $dir\n"
	"where $dic is a dictionary name,\n"
	"$text is the text file from which dictionary is generated\n"
	"If $text is omitted, lkdict shows only line addresses\n"
	"options:\n"
	"  -p $prefix	# prefix to \"dict:\", \"list:\" and \"indx:\".\n"
	"  -d $dir		# directory of the dictionary.\n"
	"  -Q			# input/output is QID format (%%016llx).\n"
	"  -n			# mark lines with the line number.\n"
	"  -H $header	# mark lines with $header.\n"
	"  -S $nsep		# string that separate line number and matched text. (default \":\")\n"
	);
	exits("usage");
}

void
tr(char *s, char from, char to)
/*	translate a char to another	*/
{
	char *p;
	p = s;
	while(*p){
		if(*p == from)
			*p = to;
		p++;
	}
}

int
lineno(vlong pos, vlong *lines, int n)
/* binary search */
{
	int low = 0;
	int high = n -1;
	int k;
	while(low <= high){
		k = (low + high)/2;
		if(pos == lines[k])
			return k;
		else
		if(pos < lines[k])
			high = k - 1;
		else
			low = k + 1;
	}
	sysfatal("lkdict: nextline %lld not found:\n %s", pos, comline);
	return -1; // not found
}


void
loutput(int fd, vlong lines[], int lcount, vlong sublines[], int count)
/*	sublines[] is a subset of lines[]	*/
{
	vlong sp,ep;
	int lnum,k,n;
	char buf[BUFSIZE];

	for(k = 0; k < count; k++){
		lnum = lineno(sublines[k],lines,lcount);
		if(seek(fd, lines[lnum], 0) < 0)
			sysfatal("seek: %r");
		sp = lines[lnum];
		ep = lines[lnum+1];
		while(BUFSIZE < ep -sp && (n = read(fd, buf, BUFSIZE-1)) > 0){
			buf[n] = 0;
			if(nflag)
				Bprint(&bout,"%s%d%s%s", header, lnum+1, nsep, buf);
			else
				Bprint(&bout,"%s%s", header, buf);
			sp += n;
		}
		n = read(fd, buf, ep - sp);
		buf[n] = 0;
		if(nflag)
			Bprint(&bout,"%s%d%s%s", header, lnum+1, nsep, buf);
		else
			Bprint(&bout,"%s%s", header, buf);
	}
}

char *
seekrd2str(int fd, vlong sp, vlong ep)
{
	char *bp;
	vlong n;
	if(seek(fd, sp, 0) < 0)
		sysfatal("seekrd2str:fseek: %r");
	bp = malloc(ep - sp + 1); // room for last '\0'
	if(!bp)
		sysfatal("seekrd2str:malloc: %r");
	n = read(fd, bp, ep - sp);
	if(n < 0)
		sysfatal("seekrd2str:read: %r");
	bp[n] = 0;
	return bp;
}


int
vbyte(char *buf, vlong v)
{
	char *b;
	b = buf;
	while(v){
		*b++ = v & 0x7f;
		v = (v >> 7);
	}
	buf[0] = 0x80 | buf[0];
	return b - buf;
}
/*	vbyte() is not used here, but it is wise not to remove.
*	created variable bytes using vbyte() is read by rdlist()	*/

#ifdef RDLIST_BIO
/*	look "RDLIST COMMENT"	*/
void
rdlist(vlong *pos, int n, vlong offset)
{
	char buf[1];
	int m,k;
	vlong sum;
	Bseek(&list, offset, 0);
	sum = 0;
	if(Bread(&list, buf, 1) != 1)
		sysfatal("list read error1: %r");
	if(*buf & 0x80 == 0)
		sysfatal("invalid list %02x", *buf);
	while(n) {
		*buf = (*buf & 0x7f);
		m = 0;
		for(k = 0;(*buf & 0x80) == 0; k++){	
			m += (*buf << 7*k);
			if(Bread(&list, buf, 1) != 1) // EOF
				break;
		};
		sum += m;
		*pos++ = sum;
		n--;
	};
}
#endif

#ifdef RDLIST_RAW
/*	look "RDLIST COMMENT"	*/
void
rdlist(vlong *pos, int n, vlong offset)
{
	/*	one vlong positive number is converted into list at most 9 bytes.
	*	therefore reading 9*n bytes for n vlong are enough
	*	in most case, n is vary small.
	*	the exception is "*" (line info).
	*	we have n as much as 2208943 (Eijiro)	*/

	char buf[8*1024];
	char *bp, *ep;
	int m,k, m0, m1;
	vlong sum;
	m0 = sizeof(buf);
	m1 = n * (1 + sizeof(vlong));
	if(m1 < m0)
		m0 = m1;

	seek(fd_list, offset, 0);
	m0 = read(fd_list, buf, m0);
	if(m0 < 0)
		sysfatal("rdlist:read:%r");
	bp = buf;
	ep = buf + m0;

	sum = 0;
	if(*bp & 0x80 == 0)
		sysfatal("rdlist:invalid list %02x", *bp);
	while(n) {
		*bp = (*bp & 0x7f);
		m = 0;
		for(k = 0;(*bp & 0x80) == 0; k++){	
			m += (*bp << 7*k);
			bp++;
			if(bp == ep){
				/*	this case should happen for larg n.
				*	for simplicity, we read list blindly to buf as much as possible	*/
				m0 = read(fd_list, buf, sizeof(buf));
				if(m0 < 0) // EOF
					break;
				bp = buf;
				ep = buf + m0;
			}
		};
		sum += m;
		*pos++ = sum;
		n--;
	};
}
#endif


#ifdef XXX
vlong
nextline(vlong pos, vlong *lines, int n)
/* for the temporally, we apply serial search */
{
	int k;
	for(k = 0; k < n; k++){
		if(lines[k] == pos)
			return lines[k+1];
	}
	sysfatal("lkdict: nextline %lld not found:\n %s", pos, comline);
	return -1; // not found
}
#endif

//#ifdef XXX
vlong
nextline(vlong pos, vlong *lines, int n)
/* binary search */
{
	int low = 0;
	int high = n -1;
	int k;
	while(low <= high){
		k = (low + high)/2;
		if(pos == lines[k])
			return lines[k+1];
		else
		if(pos < lines[k])
			high = k - 1;
		else
			low = k + 1;
	}
	sysfatal("lkdict: nextline %lld not found:\n %s", pos, comline);
	return -1; // not found
}
//#endif


int
match(Reprog *re, char *str)
/*	test if matched	*/
{
	Resub m;
	m.sp = str;
	m.ep = nil;
	if(regexec(re, str, &m, 1))
		return 1;
	return 0;
}

int
compar(int *x, int *y)
{
	return *x - *y;
}

int
uniq(vlong *a, int n)
{
	int i,j;
	if(n < 2)
		return n;
	i = 0;
	j = 1;
	while(j < n){
		if(a[i] == a[j])
			j++;
		else{
			i++;
			a[i] = a[j];
			j++;
		}
	}
	return i + 1;
}

vlong *
matchlines(int *mp, List *lroot)
{
	int m,n,k,count;
	vlong *array, offset;
	List *ls;
	Pos *d;
	m = 0;
	for(ls = lroot; ls; ls = ls->next){
		d = ls->data;
		m += d->count;
	}

	array = ecalloc(m, sizeof(vlong));
	n = 0;
	for(ls = lroot; ls; ls = ls->next){
		d = ls->data;
		count = d->count;
		offset = d->offset;
		rdlist(&array[n], count, offset);
		n += count;
	}
	qsort(array,m,sizeof(vlong), (int (*)(const void*, const void*))compar);
	DBG {
		for(k = 0; k < m; k++)
			print("%d %lld\n", k, array[k]);
	}
	m = uniq(array,m);
	DBG {
		for(k = 0; k < m; k++)
			print("%d %lld\n", k, array[k]);
	}
	*mp = m;
	return array;
}

int
common(vlong *co, vlong *a1, int n1, vlong *a2, int n2)
/*	serial common */
{
	int k,k1,k2;
	k = k1 = k2 = 0;
	while(k1 < n1 && k2 < n2){
		if(a1[k1] < a2[k2])
			k1++;
		else
		if(a1[k1] > a2[k2])
			k2++;
		else{
			co[k++] = a1[k1++];
			k2++;
		}
	}
	return k;
}

vlong *
commonlines(int *mp, List *lroot[], int nre)
{
	int k,m,m0;
	vlong *ar[MAXFIELD];
	int  na[MAXFIELD];
	vlong *co,*com;

	for(k = 0; k < nre; k++){
		ar[k] = matchlines(&m, lroot[k]);
		na[k] = m;
	}

	co = ar[0];
	m = m0 = na[0];

	if(nre == 1){
		*mp = m0;
		return co;
	}

	if(m > na[1])
		m = na[1];
	com = ecalloc(m,sizeof(vlong));
	for(k = 1; k < nre; k++){
		m0 = common(com, co, m0, ar[k], na[k]);
		co = com;
	}
	for(k = 1; k < nre; k++)
		free(ar[k]);
	
	*mp = m0;
	return com;
}

void
moutput(int m,char *args[], int nq, vlong *lines, int lcount)
{
	int n,count;
	vlong offset;
	vlong *pos;
	Pos *d;

	offset = atoll(args[1]);
	count = atoi(args[2]);
	if(nq == 1){
		pos = ecalloc(count, sizeof(vlong));
		rdlist(pos, count, offset);
		if(NoText){
			for(n = 0; n < count; n++)
				Bprint(&bout,outfmt, pos[n]);
		}
		else
			loutput(fd_text, lines, lcount, pos, count);
		free(pos);
	}
	else{
		d = ecalloc(1,sizeof(Pos));
		d->offset = offset;
		d->count = count;
		lroot[m] = Lpush(lroot[m],d);
		/*	don't free(d) here	*/
	}
}

void
Apush(Abuf *bp, vlong d)
{
	vlong *p;
	int cp;
	if(bp->array == nil){
		bp->array = ecalloc(1024,sizeof(vlong));
		bp->size = 1024;
		bp->cp = 0;
	}
	cp = bp->cp;
	if(cp == bp->size){
		p = ecalloc(2*(bp->size),sizeof(vlong));
		p = memcpy(p, bp->array, bp->size);
		free(bp->array);
		bp->array = p;
		bp->size = 2*(bp->size);
		bp->cp = cp;
	}
	bp->array[cp++] = d;
	bp->cp = cp;
}

/*	we assume input data is small */
void
rdstdin(Reprog *re[], int nq)
{
	Rdbuf *rb;
	char *p, *args[MAXFIELD];
	int k,n,m,lcount, count;
	vlong *pos;
	vlong *lines;
	vlong offset;
	Abuf *abp[MAXFIELD], *labp;

	for(k = 0; k < MAXFIELD; k++)
		abp[k] = new(Abuf);
	labp = new(Abuf);
	rb = rdinit(0, 32*1024);
	while((p = rdline(rb))){ /* assign = */
		tr(p,SPACE,TAB);
		m = split(p, args, MAXFIELD,TAB);
		if(m != 2)
			sysfatal("invalid format stdin %s", p);
		offset = atoll(args[1]);
		if(strcmp(args[0],"*") == 0){
			Apush(labp,offset);
			continue;
		}
		for(k = 0; k < nq; k++){
			if(match(re[k], args[0]))
				Apush(abp[k], offset);
		}
	}

	lcount = labp->cp;
	lines = labp->array;

	if(nq == 1){
		count = abp[0]->cp;
		if(count){
			pos = abp[0]->array;
			qsort(pos,count,sizeof(vlong), (int (*)(const void*, const void*))compar);
			count = uniq(pos,count);
			if(NoText){
				for(n=0; n < count; n++)
					Bprint(&bout,outfmt, pos[n]);
			}
			else
				loutput(fd_text, lines, lcount, pos, count);
		}
	}
	else{
		vlong *ar[MAXFIELD];
		int  m0, na[MAXFIELD];
		vlong *co,*com;
		for(k = 0; k < nq ; k++){
			count = na[k] = abp[k]->cp;
			pos = ar[k] = abp[k]->array;
			qsort(pos,count,sizeof(vlong), (int (*)(const void*, const void*))compar);
			na[k] = uniq(pos,count);

		}
		co = ar[0];
		m = m0 = na[0];
		if(m > na[1])
			m = na[1];
		com = ecalloc(m,sizeof(vlong));
		for(k = 1; k < nq; k++){
			m0 = common(com, co, m0, ar[k], na[k]);
			co = com;
			if(m0 == 0)
				break;
		}
		count = m0;
		pos = co;
		if(NoText){
			for(n=0; n < count; n++)
				Bprint(&bout,outfmt, pos[n]);
		}
		else
			loutput(fd_text, lines, lcount, pos, count);

		for(k = 1; k < nq; k++)
			free(ar[k]);
	}
	if(lines)
		free(lines);
	rdfree(rb);
}


void
rddict(int fd, Reprog *re[], int nq)
{
	Rdbuf *rb;
	char *p, *args[MAXFIELD];
	int k,n,m,lcount;
	vlong offset;
	vlong *lines, *colines;

	rb = rdinit(fd, 32*1024);
	lines = nil;
	lcount = 0;
	if(!NoText){
		p = rdline(rb);
		tr(p, SPACE,TAB);
		m = split(p, args, MAXFIELD,TAB);
		if(m != 3)
			sysfatal("invalid format dict %s",p);
		if(strcmp(args[0],"*") != 0)
			sysfatal("invalid format dict %s", args[0]);
		offset = atoll(args[1]);
		lcount = atoi(args[2]);
		lines = ecalloc(lcount, sizeof(vlong));
		rdlist(lines, lcount, offset);
	}
	while((p = rdline(rb))){ /* assign = */
		tr(p, SPACE,TAB);
		n = split(p, args, MAXFIELD,TAB);
		if(n != 3)
			sysfatal("invalid format dict %s",p);
		for(k = 0; k < nq; k++){
			if(match(re[k], args[0]))
				moutput(k,args,nq,lines,lcount);
		}
	}

	if(nq != 1){
		colines = commonlines(&m, lroot, nq);
		loutput(fd_text, lines, lcount, colines, m);
		free(colines);
	}
	if(lines)
		free(lines);
	rdfree(rb);
}

void
rdindx(int fd, Reprog *re[], char *qe[], int nq)
{
	vlong o_start, o_end;
	Rdbuf *rb;
	char *p, *s, *s0, *s1, *t, *args[MAXFIELD], *args2[MAXFIELD];
	int k,n,m,lcount;
	vlong offset;
	vlong *lines, *colines;
	char *word;

	rb = rdinit(fd, 32*1024);

	/*	reading index	*/
	lines = nil;
	lcount = 0;
	if(!NoText){
		p = rdline(rb);
		tr(p, SPACE,TAB);
		m = split(p, args, MAXFIELD,TAB);
		if(m != 3)
			sysfatal("invalid format indx %s",p);
		if(strcmp(args[0],"*") != 0)
			sysfatal("invalid format indx %s", args[0]);
	
		o_start = atoll(args[1]);
		o_end = atoll(args[2]);
		s = seekrd2str(fd_dict, o_start, o_end);
		/*	the s is a single line with '\n' at the end	*/
		p = strchr(s,'\n');
		if(p)
			*p = 0;
		tr(s, SPACE,TAB);
		m = split(s, args, MAXFIELD,TAB);
		offset = atoll(args[1]);
		lcount = atoi(args[2]);
		free(s);
		lines = ecalloc(lcount, sizeof(vlong));
		rdlist(lines, lcount, offset);
	}

	while((p = rdline(rb))){ /* assign = */
		/* the format is:
		*  word start end
		*  with separator ' ' */
		tr(p,SPACE,TAB);
		m = split(p, args, MAXFIELD,TAB);
		if(m != 3)
			sysfatal("invalid format indx %s",p);
		word = args[0];
		for(k = 0; k < nq; k++){
			s = strstr(qe[k],word);
			if(s && (s[-1] == '(' || s[-1] == '|')){
				o_start = atoll(args[1]);
				o_end = atoi(args[2]);
				s = s0 = seekrd2str(fd_dict, o_start, o_end);
				/*	the format is
				*	word pos n
				*	separated by TAB
				*/
				tr(s,SPACE,TAB);
				//print("#seekrd2str: %lld %lld\n%s",o_start, o_end,s);//DBG
				while(*s){
					s1 = strchr(s,'\n');
					t = strchr(s,TAB);
					if(t)
						*t = 0;
					//print("# %s",s);//DBG
					if(match(re[k], s)){
						if(t)
							*t = TAB;
						t = strchr(s,'\n');// should exist
						*t = 0;
						//print("# %s\n",s);//DBG
						n = split(s, args2, MAXFIELD,TAB);
						if(n != 3)
							sysfatal("invalid format indx %s",s);
						moutput(k,args2,nq,lines,lcount);
					}
					s = s1 + 1;
				}
				free(s0);
			}
		}
	}

	if(nq != 1){
		colines = commonlines(&m, lroot, nq);
		if(NoText){
			for(n = 0; n < m; n++)
				Bprint(&bout,outfmt, colines[n]);
		}
		else
			loutput(fd_text, lines, lcount, colines, m);
		free(colines);
	}
	if(lines)
		free(lines);
	rdfree(rb);
}


void
main(int argc, char *argv[])
{
	int nq,k;
	char buf[256];
	char *args[MAXFIELD];
	Reprog *re[MAXFIELD];
	char *query, *qe[MAXFIELD];

	ARGBEGIN{
	case 'p': prefix=EARGF(usage("option p"));
		break;
	case 'd': dir=EARGF(usage("option d"));
		if(dir[strlen(dir)-1] != '/')
			dir = smprint("%s/", dir);
		break;
	case 'Q':
		outfmt = "%016llx\n";
		break;
	case 'n': nflag++;
		break;
	case 'H': header=EARGF(usage("option H"));
		break;
	case 'S': nsep=EARGF(usage("option S"));
		break;
	default: usage("undefined option");
	}ARGEND

	if(argc < 2 || argc > 3)
		usage("args out of range");

	Binit(&bout, 1, OWRITE);

	if(argc==2)
		comline = smprint("lkdict [%s] %s", argv[0], argv[1]);
	else
		comline = smprint("lkdict [%s] %s %s", argv[0], argv[1], argv[2]);
	/*	comline is used for sysfatal()	*/

	query = *argv++;
	nq = split(query,args,MAXFIELD,SPACE);
	for(k = 0; k < nq; k++){
		snprint(buf,sizeof(buf),"^(%s)",args[k]);
		//print("%s\n",buf);//DBG
		qe[k] = strdup(buf);
		re[k] = regcomp(buf);
	}

	if(strcmp(*argv,"-") == 0){
		argv++;
		fd_text = open(*argv, OREAD);
		if(fd_text < 0)
			sysfatal("%s not open: %r",*argv);
		rdstdin(re, nq);
		exits(nil);
	}

	snprint(buf,sizeof(buf),"%s%sindx:%s",dir,prefix,*argv);
	fd_indx = open(buf, OREAD); // the status is checked later

	snprint(buf,sizeof(buf),"%s%sdict:%s",dir,prefix,*argv);
	fd_dict = open(buf, OREAD);
	if(fd_dict < 0)
		sysfatal("%s not open: %r", buf);

	snprint(buf,sizeof(buf),"%s%slist:%s",dir,prefix,*argv);
	fd_list = open(buf, OREAD);
	if(fd_list < 0)
		sysfatal("%s not open: %r", buf);

	argv++;
	if(*argv != nil){
		fd_text = open(*argv, OREAD);
		if(fd_text < 0)
			sysfatal("%s not open: %r",*argv);
	}
	else
		NoText = 1;

#ifdef RDLIST_BIO
	Binit(&list, fd_list, OREAD);
#endif

	if(fd_indx < 0)
		rddict(fd_dict, re, nq);
	else
		rdindx(fd_indx, re, qe, nq);
	exits(nil);
}
