/*
*
*	8.extwords1 -iohhl2 mkdict.c |sort -u  +0 -1 +1n >x
*	8.mkdict x
*	p dict:x
*	xd -1b list:x | p
*/


#include <u.h>
#include <libc.h>
#include <bio.h>
#include "misc.h"

#define MAXFIELD 2
#define MAXBUF 1024

char *prefix = "";
char *sep = " ";
Biobuf dict,list;

int base = 10;
char *path; // used in error message

void
usage(void)
{
	fprint(2,
	"usage: mkdict [-p prefix] [-F sep] file\n"
	"the default output is\n"
	"dict:file\n"
	"list:file\n"
	"options:\n"
	"  -p prefix	# put prefix to output files (default is '')\n"
	"  -F sep		# input/output field separator (default is single space)\n"
	);
	exits("usage");
}

int
vbyte(char *buf, vlong v)
/* the buf size must be 9
*  v must be positive	*/
{
	char *b;
	char *e = buf + 9;
	if(v == 0){
		buf[0] = 0x80;
		return 1;
	}
	b = buf;
	while(v && b < e){
		*b++ = v & 0x7f;
		v = (v >> 7);
	}
	buf[0] = 0x80 | buf[0];
	return b - buf;
}

void
doit(int fd)
{
	char *p, *args[MAXFIELD];
	int m,n, count;
	vlong offset, diff, off;
	char word[MAXBUF];
	char buf[9];
	char *ew;
	Rdbuf *rb;
	rb = rdinit(fd, 32*1024);
	ew = word + MAXBUF;
	word[0] = 0;
	count = 0;
	offset = 0;
	while((p = rdline(rb))){ /* assign = */
		m = split(p, args, MAXFIELD, *sep);
		if(m != 2)
			sysfatal("invalid format %s", p);
		if(strcmp(word,args[0]) == 0){
			off = strtoll(args[1], nil, base);
			diff = off - offset;
			if(diff < 0)
				sysfatal("# data error in %s: diff=%lld off=%lld offset=%lld", path, diff, off, offset);
			n = vbyte(buf, diff);
			Bwrite(&list, buf, n);
			offset = off;
			count++;
		}
		else{
			if(*word != 0)
				Bprint(&dict, "%s%d\n", sep, count);
			strecpy(word,ew,args[0]);
			offset = strtoll(args[1], nil, base);
			if(offset < 0)
				sysfatal("# data error in %s: path, offset=%lld", offset);
			Bprint(&dict,"%s%s%lld", word, sep, Boffset(&list));
			//print("# offset: %d\n", offset);
			n = vbyte(buf, offset);
			Bwrite(&list, buf, n);
			count = 1;
		}
	}
	Bprint(&dict, "%s%d\n", sep, count);
	//print("# offset: %d\n", offset);
	rdfree(rb);
}

void
main(int argc, char *argv[])
{	int fd,fd_dict, fd_list;
	char *name_dict, *name_list;
	ARGBEGIN{
	case 'F': sep = EARGF(usage());
		break;
	case 'p': prefix=EARGF(usage());
		break;
	case 'Q':
		base = 16;
		break;
	default: usage();
	}ARGEND

	if(argc != 1)
		usage();
	name_dict = smprint("%sdict:%s",prefix,argv[0]);
	name_list = smprint("%slist:%s",prefix,argv[0]);
	fd_dict = create(name_dict, OWRITE, 0644);
	if(fd_dict < 0)
		sysfatal("%s not open: %r",name_dict);
	fd_list = create(name_list, OWRITE, 0644);
	if(fd_list < 0)
		sysfatal("%s not open: %r",name_list);
	free(name_dict);
	free(name_list);

	Binit(&dict, fd_dict, OWRITE);
	Binit(&list, fd_list, OWRITE);

	fd = open(*argv, OREAD);
	if(fd < 0)
		sysfatal("%s not open: %r",*argv);
	path = *argv;
	doit(fd);
	close(fd);
	Bterm(&list);
	Bterm(&dict);
	exits(nil);
}

