/*
*	make meta index
*	the format is
*	word1 start1 start2
*	word2 start2 start3
*	...
*	where word is a word of minlen
*	start is the range in unit of bye.
*	you may this third field not need.
*	but the advantage of this format is,
*	we can use grep to extract the data.
*/

#include <u.h>
#include <libc.h>
#include <bio.h>
#include "misc.h"

#define DBG if(0)

int minlen = 2;

Biobuf bout;

int qflag;

void
usage(void)
{
	fprint(2,"usage: mktoc [-l minlen] index\n"
	"  mktoc splits database index into smaller pieces\n"
	"  index: index file which is sorted by dictionary order\n"
	"  minlen: the smallest word length in the index (default 2)\n"
	);
	exits("usage");
}

void
doit(int fd)
{
	char *p,*q;
	char p0[256];
	Rdbuf *rb;
	vlong sp0 = 0;
	vlong sp1;
	vlong n = 0;
	p0[0] = 0;
	rb = rdinit(fd, 32*1024);
	while((p = rdline(rb))){
		q = skipto(p," \t");
		if(! *q)
			sysfatal("no separator in line %s",p);
		*q = 0;
		q = skipnrune(p,minlen);
		*q = 0;
		if(p0[0] == 0)
			strecpy(p0, p0 + sizeof(p0), p);
		else
		if(strncmp(p,p0,strlen(p0)) != 0){
			/*
			*	this condition is true if p="abcd"and p0="ac"
			*	and false if p="abcd" and p0="ab"
			*	NB: the condition utflen(p) == minlen is critical.
			*	supppose minlen == 2.
			*	if we have a file
			*		ind_ab
			*		ind_cde
			*	and our request is "cd*", then both files can have
			*	words that match "cd*".
			*/

			/*	seek() is no use because we have already read data in buffer */
			sp1 = n;
			Bprint(&bout,"%s %lld %lld\n", p0, sp0, sp1);
			sp0 = sp1;
			strecpy(p0,p0 + sizeof(p0), p);
		}
		n += rdlnlen(rb) + 1; /* we should include '\n' */
	}
	sp1 = n;
	Bprint(&bout,"%s %lld %lld\n", p0, sp0, sp1);
	rdfree(rb);
}

void
main(int argc, char *argv[])
{
	int fd;
	
	ARGBEGIN{
	case 'l': minlen=atoi(EARGF(usage()));
		break;
	case 'q': qflag++;
		break;
	default: usage();
	}ARGEND

	if(*argv == nil)
		usage();

	Binit(&bout, 1, OWRITE);

	fd = open(*argv,OREAD);
	if(fd < 0)
		sysfatal("%s not open: %r",*argv);
	doit(fd);
	close(fd);
	Bterm(&bout);
	exits(nil);
}
